/*
 * Copyright (C) 1997 by the University of Kansas Center for Research,
 * Inc.  This software was developed by the Information and
 * Telecommunication Technology Center (ITTC) at the University of
 * Kansas.  Partial funding for this project was provided by Sprint. This
 * software may be used and distributed according to the terms of the GNU
 * Public License, incorporated herein by reference.  Neither ITTC nor
 * Sprint accept any liability whatsoever for this product.
 *
 * This project was developed under the direction of Dr. Douglas Niehaus.
 *
 * Authors: Shyam Pather, Balaji Srinivasan 
 *
 * Please send bug-reports/suggestions/comments to posix@ittc.ukans.edu
 *
 * Further details about this project can be obtained at
 *    http://hegel.ittc.ukans.edu/projects/posix/
 */

/* 2timer_test.c
 * 
 * This program demonstrates the use of POSIX.4 interval timers without 
 * queued signals. It simply creates a POSIX.4 interval timer that sends
 * a normal signal when it expires. The handler for this signal prints
 * a message, and increments a count. When the count reaches MAX_EXPIRATIONS,
 * the timer is deleted, and the program exits. 
 */

#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#ifdef __linux__
#include <posix_time.h>
#endif

#define TIMER1_SIGNAL SIGRTMIN
#define TIMER2_SIGNAL SIGRTMAX
#define DATA_VAL 17
#define MAX_EXPIRATIONS 5

pid_t pid;
int expirations;

void print_siginfo(siginfo_t *info)
{
  int overrun;
  timer_t t;

  printf("siginfo dump: ");
  printf("si_signo: %d, ", info->si_signo);
  printf("si_errno: %d, ", info->si_errno);
  switch (info->si_code) {
  case SI_TIMER:
    t = *(timer_t*)info->si_value.sival_ptr;
    printf("si_code: SI_TIMER, %d", t);
#ifdef __linux__
    printf(", timer id: %d", info->si_tid);
#endif
    overrun = timer_getoverrun(t);
    if (overrun) {
      printf(", overrun: %d\n", overrun);
    }
    else {
      printf("\n");
    }
    break;
  default:
    printf("si_code: %d\n", info->si_code);
    break;
  }
}


void timer1_handler(int signo)
{
	printf("handler timer1 expired: signal %d\n", signo);
	expirations++;
}

void timer1_action(int signo, siginfo_t *info, void *context)
{
	printf("action timer1 expired: signal %d\n", signo);
	print_siginfo(info);
	expirations++;
}

void timer2_handler(int signo)
{
	printf("handler timer2 expired: signal %d\n", signo);
}

void timer2_action(int signo, siginfo_t *info, void *context)
{
	printf("action timer2 expired: signal %d\n", signo);
	print_siginfo(info);
}

int main()
{
	timer_t t1, t2;
	struct sigevent sig1, sig2;
	struct itimerspec new_setting1, new_setting2, current;
	struct sigaction sa;
	sigset_t set;
	siginfo_t info;
	
	if ((pid = getpid()) < 0) {
		perror("getpid failed");
		exit(1);
	}

	expirations = 0;

	/* 
	 * Set up the signal event that will occur when the timer 
	 * expires. 
	 */	
	sig1.sigev_notify = SIGEV_SIGNAL;
	sig1.sigev_signo = TIMER1_SIGNAL;
	sig1.sigev_value.sival_ptr = &t1;

	sig2.sigev_notify = SIGEV_SIGNAL;
	sig2.sigev_signo = TIMER2_SIGNAL;
	sig2.sigev_value.sival_ptr = &t2;

	/*
	 * Initialize the timer setting structure.
	 */
	new_setting1.it_value.tv_sec = 0L;
	new_setting1.it_value.tv_nsec = 12345678L;
	new_setting1.it_interval.tv_sec = 0L;
	new_setting1.it_interval.tv_nsec = 20000000L;

	new_setting2.it_value.tv_sec = 0L;
	new_setting2.it_value.tv_nsec = 10000000L;
	new_setting2.it_interval.tv_sec = 0L;
	new_setting2.it_interval.tv_nsec = 10000000L;

	/* 
	 * Set up and install a signal handler for the signal that 
	 * the timer will send. 
	 */
#if 0
	sa.sa_handler = timer1_handler;
	sa.sa_flags = 0;
#else
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = timer1_action;
#endif
	sigemptyset(&sa.sa_mask);

	if (sigaction(TIMER1_SIGNAL, &sa, NULL)) {
		perror("sigaction failed");
		exit(1);
	}
#if 0
	sa.sa_handler = timer2_handler;
	sa.sa_flags = 0;
#else
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = timer2_action;
#endif
	sigemptyset(&sa.sa_mask);

	if (sigaction(TIMER2_SIGNAL, &sa, NULL)) {
		perror("sigaction failed");
		exit(1);
	}
	
	/*
	 * Create and set the timer.
	 */
	if (timer_create(CLOCK_REALTIME, &sig1, &t1)) {
		perror("timer_create failed");
		exit(1);
	}
	if (timer_create(CLOCK_REALTIME, &sig2, &t2)) {
		perror("timer_create failed");
		exit(1);
	}

	sigemptyset(&set);
	sigaddset(&set, TIMER1_SIGNAL);
	sigaddset(&set, TIMER2_SIGNAL);

	sigprocmask(SIG_BLOCK, &set, NULL);

	if (timer_settime(t1, 0, &new_setting1, NULL)) {
		perror("timer_settime failed");
		exit(1);
	}
	timer_gettime(t1, &current);
	printf("timer1: it_value=%ld.%09ld, it_interval=%ld.%09ld\n",
	       current.it_value.tv_sec,
	       current.it_value.tv_nsec,
	       current.it_interval.tv_sec,
	       current.it_interval.tv_nsec);
	if (timer_settime(t2, 0, &new_setting2, NULL)) {
		perror("timer_settime failed");
		exit(1);
	}
	timer_gettime(t2, &current);
	printf("timer2: it_value=%ld.%09ld, it_interval=%ld.%09ld\n",
	       current.it_value.tv_sec,
	       current.it_value.tv_nsec,
	       current.it_interval.tv_sec,
	       current.it_interval.tv_nsec);

	sleep(1);
	while (expirations < MAX_EXPIRATIONS) {
	  sigwaitinfo(&set, &info);
	  print_siginfo(&info);
	  if (info.si_signo == TIMER2_SIGNAL) expirations++;
	}
	/* 
	 * Busy wait until the timer expires MAX_EXPIRATIONS number 
	 * of times.
	 */

	/* 
	 * Delete the timer.
	 */	
	if (timer_delete(t1)) {
		perror("timer_delete failed");
		exit(1);
	}
	if (timer_delete(t2)) {
		perror("timer_delete failed");
		exit(1);
	}

	return 0;
}
